//
// Example of reflective water via GLSL programs
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#include    <glut.h>
#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "libTexture3D.h"
#include    "TypeDefs.h"
#include	"Vector4D.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include    "boxes.h"
#include    "GlslProgram.h"
#include    "Camera.h"
#include	"FrameBuffer.h"
#include	"VertexBuffer.h"
#include	"utils.h"
#include	"fpTexture.h"

#define		SIZE		16
#define		TEX_SIZE	128
#define		SCALE1		2.073
#define		SCALE2		4.235
#define		SCALE3		8.0973

Vector3D    eye   ( 0, 0, 0 );  		// camera position
unsigned	heightMap [16];
unsigned	reflectionMap;
unsigned	foamMap;
float		yaw   = 0;
float		pitch = 0;
float		roll  = 0;
float		amplitude   = 0.2;
float		amplitude2  = 0.2;
float		exposure    = 1.7;
float		curTime     = 0;

GLenum		mapFormat;				// texture format used for vertex map (32 bit)
GLenum		mapFormat2;				// texture format for HDR (16 bit)
unsigned	verticesMap;			// map with resulting vertices
unsigned	normalsMap;				// map with resulting normals

FrameBuffer	buffer  ( TEX_SIZE, TEX_SIZE );
FrameBuffer	buf2    ( 512, 512, FrameBuffer :: depth32 );
FrameBuffer	buf3    ( 256, 256, 0 );
FrameBuffer	buf4    ( 256, 256, 0 );

GlslProgram	r2vb;
GlslProgram program;
GlslProgram	toneMapProgram;
GlslProgram	xBlur;
GlslProgram yBlur;
GlslProgram	filterMap;

Camera		camera ( eye, yaw, pitch, roll );	// camera to be used

VertexBuffer	* vertexBuffer;		// vertex coordinates
VertexBuffer    * normalBuffer;		// normal buffer
VertexBuffer	* indexBuffer;		// vertex index buffer

void reshape ( int w, int h );


float	* buf = (float *) malloc ( TEX_SIZE * TEX_SIZE * 4 * sizeof ( float ) );

void	loadHeightMaps ()
{
	char	str [128];
	
	for ( int i = 0; i < 16; i++ )
	{
		sprintf ( str, "Images/WaterH%02d.jpg", i );
		
		heightMap [i] = createNormalHeightMapFromHeightMap ( true, str, 3.0 );
	}
}

void	createIndexBuffer ()
{
	int * data = (int *) malloc ( (TEX_SIZE-1)*(TEX_SIZE-1)*6*sizeof (int) );
	int	  k    = 0;
	
	for ( int i = 0; i < TEX_SIZE - 1; i++ )
		for ( int j = 0; j < TEX_SIZE - 1; j++ )
		{
			data [k]   = i + TEX_SIZE*j;				// first triangle (i,j)-(i+1,j) - (i,j+1)
			data [k+1] = i + TEX_SIZE*j + 1;
			data [k+2] = i + TEX_SIZE*j + TEX_SIZE;
			
			data [k+3] = i + TEX_SIZE*j + 1;			// second triangle (i+1,j)-(i+1,j+1) - (i,j+1)
			data [k+4] = i + TEX_SIZE*j + TEX_SIZE + 1;
			data [k+5] = i + TEX_SIZE*j + TEX_SIZE;
			
			k         += 6;
		}
		
	indexBuffer = new VertexBuffer ();
	
	indexBuffer -> bind    ( GL_ELEMENT_ARRAY_BUFFER_ARB );
	indexBuffer -> setData ( (TEX_SIZE - 1) * (TEX_SIZE - 1) * 6 * sizeof ( int ), data, GL_STATIC_DRAW );
	indexBuffer -> unbind  ();
	
	free ( data );
}

void	createVertexBuffer ()
{
	vertexBuffer = new VertexBuffer ();
	
	vertexBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	vertexBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	vertexBuffer -> unbind  ();
}

void	createNormalBuffer ()
{
	normalBuffer = new VertexBuffer ();
	
	normalBuffer -> bind    ( GL_PIXEL_PACK_BUFFER_ARB );
	normalBuffer -> setData ( buffer.getWidth () * buffer.getHeight () * 4 * sizeof ( float ), NULL, GL_DYNAMIC_DRAW );
	normalBuffer -> unbind  ();
}

//
// create vertex coordinates using r2vb
//
void	buildVertices ()
{
	curTime = 0.001f * glutGet ( GLUT_ELAPSED_TIME );		// time in seconds
	
	int		index      = ((int) (SCALE1*curTime)) & 0x0F;
	int		index2     = ((int) (SCALE2*curTime)) & 0x0F;
	int		index3     = ((int) (SCALE3*curTime)) & 0x0F;
	GLenum	buffers [] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };

														// load source image map
	glActiveTextureARB ( GL_TEXTURE0_ARB );
    glBindTexture      ( GL_TEXTURE_2D, heightMap [index] );
	glActiveTextureARB ( GL_TEXTURE1_ARB );
    glBindTexture      ( GL_TEXTURE_2D, heightMap [index2] );
	glActiveTextureARB ( GL_TEXTURE2_ARB );
    glBindTexture      ( GL_TEXTURE_2D, heightMap [index3] );
	glActiveTextureARB ( GL_TEXTURE3_ARB );
    glBindTexture      ( GL_TEXTURE_2D, heightMap [(index+1) & 0x0F] );
	glActiveTextureARB ( GL_TEXTURE4_ARB );
    glBindTexture      ( GL_TEXTURE_2D, heightMap [(index2+1) & 0x0F] );
	glActiveTextureARB ( GL_TEXTURE5_ARB );
    glBindTexture      ( GL_TEXTURE_2D, heightMap [(index3+1) & 0x0F] );
	glActiveTextureARB ( GL_TEXTURE0_ARB );

	r2vb.bind      ();									// perform rendering to FBO texture
	buffer.bind    ();
	glDrawBuffers  ( 2, buffers );
	startOrtho     ( TEX_SIZE, TEX_SIZE );
	drawQuad       ( TEX_SIZE, TEX_SIZE );
	endOrtho       ();
	r2vb.unbind    ();
	
														// copy vertex data from texture to vertex buffer
	glReadBuffer           ( GL_COLOR_ATTACHMENT0_EXT );
	vertexBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
	glReadPixels           ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, NULL );
	vertexBuffer -> unbind ();

	glReadBuffer           ( GL_COLOR_ATTACHMENT1_EXT );
	normalBuffer -> bind   ( GL_PIXEL_PACK_BUFFER_ARB );
	glReadPixels           ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, NULL );
	normalBuffer -> unbind ();

	glReadPixels ( 0, 0, TEX_SIZE, TEX_SIZE, GL_RGBA, GL_FLOAT, buf );
	
	buffer.unbind          ();
}

void displayBoxes ()
{
	int		index = ((int) (curTime+12)) & 0x0F;

    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glActiveTextureARB ( GL_TEXTURE0_ARB );

	drawCubicMappedBox ( Vector3D ( SIZE, SIZE, SIZE ), reflectionMap );

	glActiveTextureARB ( GL_TEXTURE3_ARB );
	glBindTexture      ( GL_TEXTURE_2D, foamMap );
	glActiveTextureARB ( GL_TEXTURE2_ARB );
	glBindTexture      ( GL_TEXTURE_2D, heightMap [(index+1)&0x0F] );
	glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, heightMap [index] );
    glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_CUBE_MAP_ARB, reflectionMap );
	
	glPopMatrix  ();
    glPushMatrix ();

    glPushClientAttrib   ( GL_CLIENT_VERTEX_ARRAY_BIT );
	
	glEnableClientState  ( GL_VERTEX_ARRAY );
	vertexBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glVertexPointer      ( 4, GL_FLOAT, 0, NULL );
	
	glEnableClientState  ( GL_TEXTURE_COORD_ARRAY );
	normalBuffer -> bind ( GL_ARRAY_BUFFER_ARB );
	glTexCoordPointer    ( 4, GL_FLOAT, 0, NULL );
	
	indexBuffer  -> bind ( GL_ELEMENT_ARRAY_BUFFER_ARB );
	
	program.bind   ();
	glDrawElements ( GL_TRIANGLES, (TEX_SIZE-1)*(TEX_SIZE-1)*6, GL_UNSIGNED_INT, 0 );
	program.unbind ();

	vertexBuffer -> unbind ();
	normalBuffer -> unbind ();
	indexBuffer  -> unbind ();
	
	glPopClientAttrib ();

	glActiveTextureARB ( GL_TEXTURE2_ARB );
	glBindTexture      ( GL_TEXTURE_2D, 0 );
	glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, 0 );
    glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_CUBE_MAP_ARB, 0 );
	glDisable          ( GL_TEXTURE_CUBE_MAP_ARB );
	glDisable          ( GL_TEXTURE_RECTANGLE_ARB );
	
    glPopMatrix ();
}

void	blurMap ()
{
	startOrtho ( buf3.getWidth (), buf3.getHeight () );
	
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buf2.getColorBuffer () );
	
	buf3.bind ();
	filterMap.bind ();
	drawQuad ( buf3.getWidth (), buf3.getHeight () );
	filterMap.unbind ();
	buf3.unbind ();
	
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buf3.getColorBuffer () );
	
	buf4.bind ();
	xBlur  .bind ();									// perform x-blurring first
	drawQuad ( buf4.getWidth (), buf4.getHeight () );
	xBlur.unbind   ();
	buf4.unbind ();
	
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buf4.getColorBuffer () );
	
	buf3.bind ();
	yBlur.bind ();
	drawQuad ( buf3.getWidth (), buf3.getHeight () );
	yBlur.unbind ();
	buf3.unbind ();
														// now in buffer3 we have blurred FP map
	endOrtho ();
}

void display ()
{
	buildVertices ();

	reshape ( buf2.getWidth (), buf2.getHeight () );
	
	camera.apply ();

	glMatrixMode ( GL_TEXTURE );
	glPushMatrix ();
	glRotatef    ( 90, 1, 0, 0 );

	glMatrixMode ( GL_MODELVIEW );

	buf2.bind ();
	
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    displayBoxes ();

	buf2.unbind ();

	glMatrixMode   ( GL_TEXTURE );
	glLoadIdentity ();

	glMatrixMode ( GL_MODELVIEW );

	blurMap ();

	startOrtho ( buf2.getWidth (), buf2.getHeight () );
	
	glClear ( GL_COLOR_BUFFER_BIT );

	buf3.buildMipmaps ();
	
	glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buf3.getColorBuffer () );
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buf2.getColorBuffer () );
	glEnable           ( GL_TEXTURE_2D );
	
	toneMapProgram.bind   ();
	drawQuad       ( buf2.getWidth (), buf2.getHeight () );
	toneMapProgram.unbind ();
	
	endOrtho ();
	
    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
	camera.setViewSize ( w, h, 60 );
	camera.apply       ();
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );

   	if ( key == 'w' || key == 'W' )
   		camera.moveBy ( camera.getViewDir () * 0.2 );
   	else
   	if ( key == 'x' || key == 'X' )
   		camera.moveBy ( -camera.getViewDir () * 0.2 );
   	else
   	if ( key == 'a' || key == 'A' )
   		camera.moveBy ( -camera.getSideDir () * 0.2 );
   	else
   	if ( key == 'd' || key == 'D' )
   		camera.moveBy ( camera.getSideDir () * 0.2 );
	else
	if ( key == '+' )
		amplitude += 0.02;
	else
	if ( key == '-' )
		amplitude -= 0.02;
	else
	if ( key == '*' )
		amplitude2 += 0.02;
	else
	if ( key == '/' )
		amplitude2 -= 0.02;
	else
	if ( key == '[' )
		exposure -= 0.025;
	else
	if ( key == ']' )
		exposure += 0.025;
		
	glutPostRedisplay ();
}

void    specialKey ( int key, int x, int y )
{
    if ( key == GLUT_KEY_UP )
        yaw += M_PI / 90;
    else
    if ( key == GLUT_KEY_DOWN )
        yaw -= M_PI / 90;
	else
    if ( key == GLUT_KEY_RIGHT )
        roll += M_PI / 90;
    else
    if ( key == GLUT_KEY_LEFT )
        roll -= M_PI / 90;

	camera.setEulerAngles ( yaw, pitch, roll );

    glutPostRedisplay ();
}

void	mouseFunc ( int x, int y )
{
	static	int	lastX = -1;
	static	int	lastY = -1;

	if ( lastX == -1 )				// not initialized
	{
		lastX = x;
		lastY = y;
	}

	yaw  -= (y - lastY) * 0.02;
	roll += (x - lastX) * 0.02;

	lastX = x;
	lastY = y;

	camera.setEulerAngles ( yaw, pitch, roll );

	glutPostRedisplay ();
}

void    animate ()
{
	static	float	lastTime = 0.0;
	float			time     = 0.001f * glutGet ( GLUT_ELAPSED_TIME );

    lastTime = time;

    program.bind ();
    program.setUniformVector ( "eyePos",      Vector4D ( camera.getPos (), 1 ) );
    program.setUniformFloat  ( "time",        time );
    program.setUniformFloat  ( "amplitude",   amplitude );
    program.setUniformFloat  ( "amplitude2",  amplitude2 );
    program.unbind ();
	
	r2vb.bind            ();
    r2vb.setUniformFloat ( "time",        time );
    r2vb.setUniformFloat ( "amplitude",   amplitude );
	r2vb.unbind          ();

	toneMapProgram.bind            ();
    toneMapProgram.setUniformFloat ( "exposure",    exposure );
	toneMapProgram.unbind          ();
	
    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    glutInitWindowSize  ( 512, 512 );


                                // create window
    glutCreateWindow ( "OpenGL reflective water with precomputed heights and HDR" );

                                // register handlers
    glutDisplayFunc       ( display    );
    glutReshapeFunc       ( reshape    );
    glutKeyboardFunc      ( key        );
    glutSpecialFunc       ( specialKey );
    glutIdleFunc          ( animate    );
    glutPassiveMotionFunc ( mouseFunc  );

    init           ();
    initExtensions ();

	assertExtensionsSupported ( "ARB_pixel_buffer_object" );

	if ( (mapFormat = fpFormatWithPrecision ( 32 ) ) == 0 )
	{
		printf ( "Floating-point textures not supported !\n" );
		
		return 1;
	}
	
	if ( (mapFormat2 = fpFormatWithPrecision ( 16 ) ) == 0 )
	{
		printf ( "Floating-point textures not supported !\n" );
		
		return 1;
	}
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported.\n" );

		return 1;
	}
	
    if ( !program.loadShaders ( "water-6.vsh", "water-6.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", program.getLog ().c_str () );

        return 2;
    }

    if ( !r2vb.loadShaders ( "r2vb-2.vsh", "r2vb-2.fsh" ) )
    {
        printf ( "Error loading shaders:\n%s\n", r2vb.getLog ().c_str () );

        return 3;
    }

    if ( !toneMapProgram.loadShaders ( "toneMap.vsh", "toneMap-5.fsh" ) )
    {
        printf ( "Error loading toneMap shaders:\n%s\n", toneMapProgram.getLog ().c_str () );

        return 3;
    }
	
    if ( !xBlur.loadShaders ( "b-x-3.vsh", "b-x-3.fsh" ) )
    {
        printf ( "Error loading xBlur shaders:\n%s\n", xBlur.getLog ().c_str () );

        return 3;
    }

    if ( !yBlur.loadShaders ( "b-y-3.vsh", "b-y-3.fsh" ) )
    {
        printf ( "Error loading yBlur shaders:\n%s\n", yBlur.getLog ().c_str () );

        return 3;
    }
	
    if ( !filterMap.loadShaders ( "filterMap.vsh", "filterMap.fsh" ) )
    {
        printf ( "Error loading filterMap shaders:\n%s\n", yBlur.getLog ().c_str () );

        return 3;
    }
	

	const char * faces [6] =
	{
		"../../Textures/Cubemaps/skyrt.bmp",
		"../../Textures/Cubemaps/skylf.bmp",
		"../../Textures/Cubemaps/skydn.bmp",
		"../../Textures/Cubemaps/skyup.bmp",
		"../../Textures/Cubemaps/skybk.bmp",
		"../../Textures/Cubemaps/skyfr.bmp",
	};

	loadHeightMaps ();
	reflectionMap = createCubeMap   ( true, faces );
	foamMap       = createTexture2D ( true, "Images/Foam.jpg" );

    r2vb.bind ();
	r2vb.setTexture ( "h1Map",  0 );
	r2vb.setTexture ( "h2Map",  1 );
	r2vb.setTexture ( "h3Map",  2 );
	r2vb.setTexture ( "h1Map2", 3 );
	r2vb.setTexture ( "h2Map2", 4 );
	r2vb.setTexture ( "h3Map2", 5 );
    r2vb.unbind ();

    program.bind ();
    program.setTexture ( "envMap",  0 );
    program.setTexture ( "h4Map",   1 );
    program.setTexture ( "h4Map2",  2 );
	program.setTexture ( "foamMap", 3 );
    program.unbind ();

	xBlur.bind              ();
	xBlur.setTexture        ( "mainTex", 0 );
	xBlur.unbind            ();
	yBlur.bind              ();
	yBlur.setTexture        ( "mainTex", 0 );
	yBlur.unbind            ();
	
	filterMap.bind       ();
	filterMap.setTexture ( "mainMap", 0 );
	filterMap.unbind     ();
	
	toneMapProgram.bind       ();
	toneMapProgram.setTexture ( "mainMap", 0 );
	toneMapProgram.setTexture ( "blurMap", 1 );
	toneMapProgram.unbind     ();
	
	camera.setRightHanded ( false );

	buffer.create ();

	unsigned vertMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	unsigned normMap = buffer.createColorRectTexture ( GL_RGBA, mapFormat );
	
	buffer.bind   ();
	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, vertMap, 0 ) )
		printf ( "buffer error with color attachment 0\n");

	if ( !buffer.attachColorTexture ( GL_TEXTURE_RECTANGLE_ARB, normMap, 1 ) )
		printf ( "buffer error with color attachment 1\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );

	buffer.unbind ();
		
	unsigned	map2 = buf2.createColorTexture ( GL_RGBA, mapFormat2, GL_CLAMP );

	buf2.create ();
	buf2.bind   ();
	buf2.attachColorTexture ( GL_TEXTURE_2D, map2, 0 );
	buf2.unbind ();
	
	if ( !buf2.isOk () )
		printf ( "Error with buf2\n" );

	unsigned	map3 = buf3.createColorTexture ( GL_RGBA, mapFormat2, GL_CLAMP );
	
	buf3.create ();
	buf3.bind ();
	buf3.attachColorTexture ( GL_TEXTURE_2D, map3, 0 );
	buf3.unbind ();
	
	if ( !buf3.isOk () )
		printf ( "Error with buf3\n" );

	unsigned	map4 = buf4.createColorTexture ( GL_RGBA, mapFormat2, GL_CLAMP );
	
	buf4.create ();
	buf4.bind ();
	buf4.attachColorTexture ( GL_TEXTURE_2D, map4, 0 );
	buf4.unbind ();
	
	if ( !buf4.isOk () )
		printf ( "Error with buf4\n" );


	createVertexBuffer ();
	createNormalBuffer ();
	createIndexBuffer  ();
	
	printf ( "Use:\n\t+ and - keys to control vertex wave ampliture\n\t* and / to control EMBM\n\t[ and ] to control exposure\n\t WXAD to move camera\n" );
	
    glutMainLoop ();

    return 0;
}
